"""
This addon
  saves all signal connections of a selected robot to a json file
  loads all signal connections from a json file to a selected robot

28.03.2023 - V 1.0 - Initial version

"""

import os
import sys
import json
from vcCommand import *
from vcHelpers.Selection import *
from vcHelpers.Application import *
from vcApplication import *
import kuka_sim_extensions.Libs.DS_Library.ds_sim_helpers as ds_helpers

GLOBAL_ERROR = "ERROR - SAVE/LOAD SIGNAL CONNECTION ADDON"

def first_state():
  reload(ds_helpers)
  global save_button, load_button, path, help_button
  """ Set up task pane for UI in KUKA.Sim """
  groupNr = 0  # groupNr is used to organize displayed elements in the task pane which is shown by executeInActionPanel()

  save_button = createProperty(VC_BUTTON, 'SAVE signal connections of selected robot')
  save_button.IsVisible = True
  save_button.OnChanged = save_connections
  save_button.Group = groupNr
  groupNr += 1

  load_button = createProperty(VC_BUTTON, 'LOAD signal connections to selected robot')
  load_button.IsVisible = True
  load_button.OnChanged = load_connections
  load_button.Group = groupNr
  groupNr += 1

  empty_line = createProperty(VC_BUTTON, '----------------------------------------------------------------------------------------------------------------------------------------------------')
  empty_line.IsVisible = True
  empty_line.WritableWhenConnected = False
  empty_line.WritableWhenDisconnected = False
  empty_line.Group = groupNr
  groupNr += 1

  help_button = createProperty(VC_BUTTON, 'Open Help File')
  help_button.IsVisible = True
  help_button.OnChanged = open_help
  help_button.Group = groupNr
  groupNr += 1

  # display task pane in KUKA.Sim UI with all previously defined variables (prop1, prop2, etc.)
  executeInActionPanel()


def get_connections_dict(rob_signal_map):
  out_map_dict = {}
  list_connected_ports = rob_signal_map.getAllConnectedPorts()

  for port in list_connected_ports:
    list_ext_sig = rob_signal_map.getConnectedExternalSignals(port)
    comp_and_sig_list = []

    for signal in list_ext_sig:
      comp_and_sig_list.append((signal.Component.Name, signal.Name))
  
    out_map_dict[port] = comp_and_sig_list  
  
  return out_map_dict

def get_path():
  """get script path and cut the leading first 8 chars
    file:///C:\Users\sturma\OneDrive - KUKA AG\Dokumente\KUKA\KUKA.Sim 4.1\My Commands\SaveLoadSignalConnections
    ->
    C:\Users\sturma\OneDrive - KUKA AG\Dokumente\KUKA\KUKA.Sim 4.1\My Commands\SaveLoadSignalConnections
  """

  script_path = os.path.dirname(__file__)
  script_path_list = list(script_path)
  script_path_list[0:8] = ""
  script_path = "".join(script_path_list)

  return script_path


def deunicodify_hook(pairs):
  """
  This function has been copy+pasted from stackoverflow. It deals with encoding problems from unicode, 
  while importing a dict from a json file.
  https://stackoverflow.com/questions/956867/how-to-get-string-objects-instead-of-unicode-from-json/34796078#34796078

  :param pairs: pair(key, value) from a decoded JSON object that delivers an ordered list of pairs
  """
  new_pairs = []
  for key, value in pairs:
      if isinstance(value, unicode):
          value = value.encode('utf-8')
      if isinstance(key, unicode):
          key = key.encode('utf-8')
      new_pairs.append((key, value))
  return dict(new_pairs)


def save_connections(para):
  """ Saves input and output map of selected robot in two files. """

  rob = ds_helpers.get_selected_robot()
  if not rob:
    return

  rob_output_map = rob.findBehavioursByType(VC_ROBOTEXECUTOR)[0].DigitalOutputSignals
  rob_input_map = rob.findBehavioursByType(VC_ROBOTEXECUTOR)[0].DigitalInputSignals

  out_map_dict = get_connections_dict(rob_output_map)
  in_map_dict = get_connections_dict(rob_input_map)

  script_path = get_path()

  # json.dump(out_map_dict, open(script_path + "\saved_output_connections.json", 'w'))
  # It is good practice to use the with keyword when dealing with file objects. 
  # This has the advantage that the file is properly closed after its suite finishes, even if an exception is raised on the way. 
  # It is also much shorter than writing equivalent try-finally blocks.
  with open(script_path + '\saved_output_connections.json', 'w') as dump_file:
    json.dump(out_map_dict, dump_file)

  # json.dump(in_map_dict, open(script_path + '\saved_input_connections.json', 'w'))
  with open(script_path + '\saved_input_connections.json', 'w') as dump_file:
    json.dump(in_map_dict, dump_file)

  print ('Connections saved')

def connect_signals(app, rob_map, io_map_dict):
  for port in io_map_dict:
    for comp_and_signal in io_map_dict[port]:
      comp_name = comp_and_signal[0]
      signal_name = comp_and_signal[1]
      port_nr = int(port)
      destination_comp = app.findComponent(comp_name)
      
      # check if destination comp is a robot, if so we have to find the destination port via the signal map
      if destination_comp.findBehavioursByType(VC_ROBOTCONTROLLER):
        signal_name_elements = signal_name.split('_') # "Inputs_Port_1" -> example for signal_name of connected signal at port 1 of a robot:
        destination_port_nr = int(signal_name_elements[-1])
        if 'input' in signal_name.lower(): 
          destination_input_map = destination_comp.findBehaviour('Inputs')
          rob_map.connect(port_nr, destination_input_map, destination_port_nr)
        elif 'output' in signal_name.lower():
          destination_output_map = destination_comp.findBehaviour('Outputs')
          rob_map.connect(port_nr, destination_output_map, destination_port_nr)
      # if  the destination comp is not a robot, we can find the destination signal via the name ob the behaviour
      else:
        destination_signal = destination_comp.findBehaviour(signal_name)
        rob_map.connect(port_nr, destination_signal)   

def load_connections(para):
  """ Loads input and output map for selected robot from two files. """

  app = getApplication()
  rob = ds_helpers.get_selected_robot()
  if not rob:
    return

  load_path = get_path()

  outputs_file = load_path + "\saved_output_connections.json"
  inputs_file = load_path + "\saved_input_connections.json"

  # out_map_dict = json.load(open(outputs_file, 'r'))
  # It is good practice to use the with keyword when dealing with file objects. 
  # This has the advantage that the file is properly closed after its suite finishes, even if an exception is raised on the way. 
  # It is also much shorter than writing equivalent try-finally blocks.
  with open(outputs_file, 'r') as read_file:
    out_map_dict = json.load(read_file, object_pairs_hook=deunicodify_hook) # see Pyhton documentation regarding object_pairs_hook

  rob_output_map = rob.findBehavioursByType(VC_ROBOTEXECUTOR)[0].DigitalOutputSignals
  if rob_output_map:
    connect_signals(app, rob_output_map, out_map_dict)
  else:
    print ('Unable to find output map of selected robot. Aborting command.')
    return

  # in_map_dict = json.load(open(inputs_file, 'r'))
  with open(inputs_file, 'r') as read_file:
    in_map_dict = json.load(read_file, object_pairs_hook=deunicodify_hook) # see Pyhton documentation regarding object_pairs_hook

  rob_input_map = rob.findBehavioursByType(VC_ROBOTEXECUTOR)[0].DigitalInputSignals
  connect_signals(app, rob_input_map, in_map_dict)

  print ('Connections loaded')


def open_help(para):
  """
  Opens a help file for the add on.
  """
  file_path = get_path()
  os.startfile(file_path + '\\Help_File_SaveLoadSignalConnections.pdf')
  print('Open help file from ' + file_path)

# this executes the first_state method when the command is called from KUKA.Sim -> UI is opened
addState(first_state)
