"""
This script imports values for tools, bases and XHOME to a robot in the simulation from a given config.dat

08.06.2022 - prototype (derived from legacy script)
15.11.2022 - V 1.1 - added support for to import XHOME also for robots with less than 6 axis
08.03.2023 - V 1.2 - added support for uncommon config.dat syntax where X, Y, Z, A, B, C is in minor case
28.03.2023 - V 1.3 - added support for uncommon config.dat syntax where E6AXIS XHOME1={a1 0.0,a2 -90.0, ...} is in minor case
25.04.2023 - V 1.4 - updated error messages, robot selection simplified
19.02.2024 - V 1.5 - beautifying (imports etc.)
17.06.2024 - V 1.6 - enhanced support for various syntax in config.dat for HOME positions (with and without DECL GLOBAL), beatifying

"""
from vcScript import * # import of stub file to enable auto completion for VC modules (e.g. C:\Python27\Lib\vcScript if Python27 interpreter is used)
import os
import re
import sys
import vcVector # type: ignore
import vcMatrix # type: ignore
from vcHelpers.Selection import *
from itertools import tee
from kuka import * 
import kuka_sim_extensions.Libs.DS_Library.ds_sim_helpers as ds_helpers

def first_state():
  # reload(ds_helpers) # type: ignore
  # print('EXECUTING SCRIPT:\n' + __file__) # debug print
  global help_button
  
  # Set up task pane for UI in KUKA.Sim
  groupNr = 0  # groupNr is used to organize displayed elements in the task pane which is shown by executeInActionPanel()
  help_button = cmd.createProperty(VC_BUTTON, 'Open Help File')
  help_button.IsVisible = True
  help_button.OnChanged = open_help
  help_button.Group = groupNr
  groupNr += 1
  
  # display task pane in KUKA.Sim UI with all previously defined variables (prop1, prop2, etc.)
  cmd.executeInActionPanel()

  rob = ds_helpers.get_selected_robot()

  if not rob:
    print('Please select a standard KUKA robot before executing the Import Tools/Bases command.')
    return

  rsls = rob.findBehavioursByType(VC_ROBOTCONTROLLER)
  if len(rsls) == 1:
    controller = 'Robot'
    print("Robot found")
    print(rsls[0].Name)

  if len(rsls) > 1:
    print("Multiple VC_ROBOTCONTROLLER behaviors found in the component. Aborting command. Please select a standard KUKA robot to import a $config.dat")
    return

  opencmd = getApplication().findCommand("dialogOpen")
  uri = ""
  ok = True
  data_type = "Import $config.dat (*.dat)|*.dat"
  opencmd.execute(uri, ok, data_type)

  if not opencmd.Param_2:
    print("No $config.dat file selected for uploading, aborting command")
    return
  uri = opencmd.Param_1
  filename = uri[8:len(uri)]
  try:
    infile = open(filename, "r")
  except:
    print("[1] Cannot open file \'%s\' , aborting command. Check if file path is too long or contains not supported characters." % filename)
    return
  filestring = infile.read()
  infile.close()
  print("------------------------------------------------------------------------------------------------------------------------------------------------------")
  print("$config.dat File loaded from %s" %uri)
  print("------------------------------------------------------------------------------------------------------------------------------------------------------")

  '''Import Base transformation values'''
  num = r'\s*-?\d+\.?\d*[eE]?-?\d*'
  base_container_bases = rob.findBehavioursByType(VC_ROBOTCONTROLLER)[0].Bases
  BasesCount = len(base_container_bases) #128
  for i in range(BasesCount):
    curBase = str(i + 1)
    base = r'[^;][^; ]BASE_DATA\['+curBase+r'\]={X (?P<X>'+num+'),Y (?P<Y>'+num+'),Z (?P<Z>'+num+'),A (?P<A>'+num+'),B (?P<B>'+num+'),C (?P<C>'+num+')}.*'
    # [^;][^; ] Is not allowed to Start with; and ;(Freespace)
    # \ Escape before [, ] "
    # '+curBase+r' uses a variable before a number

    # for unknown reasons the syntax of X, Y, Z, A, B, C in the config.dat sometimes changes to minor case
    # standard syntax:                 BASE_DATA[1]={X 0.0,Y 0.0,Z 0.0,A 0.0,B 0.0,C 0.0}
    # uncommon but still valid syntax: BASE_DATA[1]={x 0.0,y 0.0,z 0.0,a 0.0,b 0.0,c 0.0}
    # to tackle this re.IGNORECASE is used
    re_base = re.compile(base, re.IGNORECASE)
    bases = re_base.finditer(filestring)
    base_dummy = vcMatrix.new()
    for b in bases:
      m = vcMatrix.new()
      m.P = vcVector.new(eval(b.group('X')),eval(b.group('Y')),eval(b.group('Z')))
      m.WPR = vcVector.new(eval(b.group('C')),eval(b.group('B')),eval(b.group('A')))
      base_dummy = m
      print("Imported BASE_DATA[%s] X  %f  Y  %f  Z  %f  A  %f  B  %f  C  %f" % (curBase, eval(b.group('X')), eval(b.group('Y')), eval(b.group('Z')), eval(b.group('A')), eval(b.group('B')), eval(b.group('C'))))

    # setting values also for base frames
    try:
      base_container_bases[i].PositionMatrix = base_dummy
    except IndexError:
      pass # if there are no base frames, just passing over

  print("------------------------------------------------------------------------------------------------------------------------------------------------------")

  '''Import Base names''' 
  for i in range(BasesCount):
    curBase = str(i + 1)
    basename = r'[^;][^; ]BASE_NAME\['+curBase+r'.\]=.(?P<Bname>[^"]*).'

    re_basename = re.compile(basename)
    basenames = re_basename.finditer(filestring)
    for bn in basenames:
      if bn.group('Bname') != " ":
        base_props = base_container_bases[i].Properties
        name = None
        for prop in base_props:
          if prop.Name == "Name":
            name = prop
            break
        base_name = str(bn.group('Bname'))   
        name.Value = base_name
        print("Imported BASE_NAME [%s] %s" % (curBase, base_name))
  print("------------------------------------------------------------------------------------------------------------------------------------------------------")

  '''Import Base type''' 
  for i in range(BasesCount):
    curBase = str(i+1)
    base_type = r'[^;][^; ]BASE_TYPE\['+curBase+r']=#(?P<Btype>.*)'

    re_base_type = re.compile(base_type)
    types = re_base_type.finditer(filestring)
    for type in types:
      if True: # type.group('Btype') != "NONE":
        base_props = base_container_bases[i].Properties
        interpol_type = None
        for prop in base_props:
          if prop.Name == "InterpolationMode":
            interpol_type = prop
            break
        b_type = str(type.group('Btype'))   
        if b_type == "NONE":
          interpol_type.Value = 0
        elif b_type == "BASE":
          interpol_type.Value = 1
        elif b_type == "TOOL":
          interpol_type.Value = 2
        print("Imported BASE_TYPE [%s] %s" % ( curBase, b_type))

  print("------------------------------------------------------------------------------------------------------------------------------------------------------")

  '''Import Tool transformation values'''
  num = r'\s*-?\d+\.?\d*[eE]?-?\d*'
  tool_container_tools = rob.findBehavioursByType(VC_ROBOTCONTROLLER)[0].Tools
  tool_count = len(tool_container_tools)
  for i in range(tool_count):
    cur_tool = str(i+1)
    tool = r'[^;][^; ]TOOL_DATA\['+cur_tool+r'\]={X (?P<X>'+num+'),Y (?P<Y>'+num+'),Z (?P<Z>'+num+'),A (?P<A>'+num+'),B (?P<B>'+num+'),C (?P<C>'+num+')}.*'
    # [^;][^; ] Is not allowed to Start with; and ;(Freespace)
    # \ Escape before [, ] "
    # '+cur_tool+r' uses a variable before a number

    # for unknown reasons the syntax of X, Y, Z, A, B, C in the config.dat sometimes changes to minor case
    # standard syntax:                 TOOL_DATA[1]={X 0.0,Y 0.0,Z 0.0,A 0.0,B 0.0,C 0.0}
    # uncommon but still valid syntax: TOOL_DATA[1]={x 0.0,y 0.0,z 0.0,a 0.0,b 0.0,c 0.0}
    # to tackle this re.IGNORECASE is used
    re_tool = re.compile(tool, re.IGNORECASE) # re.IGNORECASE fixes problems if config.dat syntax is in minor case
    tools = re_tool.finditer(filestring)
    tool_dummy = vcMatrix.new()
    for t in tools:
      m = vcMatrix.new()
      m.P = vcVector.new(eval(t.group('X')),eval(t.group('Y')),eval(t.group('Z')))
      m.WPR = vcVector.new(eval(t.group('C')),eval(t.group('B')),eval(t.group('A')))
      tool_dummy = m
      print("Imported TOOL_DATA[%s] X  %f  Y  %f  Z  %f  A  %f  B  %f  C  %f" % (cur_tool, eval(t.group('X')), eval(t.group('Y')), eval(t.group('Z')), eval(t.group('A')), eval(t.group('B')), eval(t.group('C'))))

    # setting values also for tool matrix
    try:
      tool_container_tools[i].PositionMatrix = tool_dummy
    except IndexError:
      pass # if there are no tool frames, just passing over

  print("------------------------------------------------------------------------------------------------------------------------------------------------------")

  '''Import Tool names''' 
  for i in range(tool_count):
    cur_Tool = str(i + 1)
    tool_name = r'[^;][^; ]TOOL_NAME\['+cur_Tool+r'.\]=.(?P<Tname>[^"]*).'

    re_tool_name = re.compile(tool_name)
    tool_names = re_tool_name.finditer(filestring)

    for tn in tool_names:
      if tn.group('Tname') != " ": # this condition allows to only execute/print not empty tool names
        tool_props = tool_container_tools[i].Properties
        description = None
        name = None
        for prop in tool_props:
          if prop.Name == 'Name':
            name = prop
            break
        tool_name = str(tn.group('Tname'))
        name.Value = tool_name
        print("Imported TOOL_NAME [%s] %s" % ( cur_Tool, tool_name ))
  
  print("------------------------------------------------------------------------------------------------------------------------------------------------------")
  
  '''Import Tool type''' 
  for i in range(tool_count):
    cur_Tool = str(i + 1)
    tool_type = r'[^;][^; ]TOOL_TYPE\['+cur_Tool+r']=#(?P<Ttype>.*)'
  
    re_tool_type = re.compile(tool_type)
    types = re_tool_type.finditer(filestring)
  
    for type in types:
      if True: # type.group('Btype') != "NONE": # this condition allows to only types different from #NONE
        tool_props = tool_container_tools[i].Properties
        interpol_type = None
        for prop in tool_props:
          if prop.Name == "InterpolationMode":
            interpol_type = prop
            break
        t_type = str(type.group('Ttype'))   
        if t_type == "NONE":
          interpol_type.Value = 0 # 0 = NONE in KUKA.Sim
        elif t_type == "BASE":
          interpol_type.Value = 1 # 1 = BASE in KUKA.Sim
        elif t_type == "TOOL":    
          interpol_type.Value = 2 # 2 = TOOL in KUKA.Sim
        print("Imported TOOL_TYPE [%s] %s" % ( cur_Tool, t_type))
  
  print("------------------------------------------------------------------------------------------------------------------------------------------------------")
  
  '''Import LOAD DATA'''
  num = r'\s*-?\d+\.?\d*[eE]?-?\d*'
  tool_container_tools = rob.findBehavioursByType(VC_ROBOTCONTROLLER)[0].Tools
  tool_count = len(tool_container_tools)
  for i in range(tool_count):
    # print "Import Load Data %s" % (i+1)
    cur_tool = str(i+1)
    # LOAD_DATA[1]={M 11.5000,CM {X 8.39629555,Y -23.5680313,Z 129.835,A 167.301086,B -85.6301880,C 102.794296},J {X 0.0588060,Y 0.0840960,Z 0.370812}}
    load_expr = r'[^;][^; ]LOAD_DATA\['+cur_tool+r'\]={M (?P<M>'+num+'),CM {X (?P<X>'+num+'),Y (?P<Y>'+num+'),Z (?P<Z>'+num+'),A (?P<A>'+num+'),B (?P<B>'+num+'),C (?P<C>'+num+')},J {X (?P<JX>'+num+'),Y (?P<JY>'+num+'),Z (?P<JZ>'+num+')}}.*'
    # [^;][^; ] Is not allowed to Start with; and ;(Freespace)
    # \ Escape before [, ] "
    # '+cur_tool+r' uses a variable before a number
    
    # for unknown reasons the syntax of X, Y, Z, A, B, C in the config.dat sometimes changes to minor case
    # to tackle this re.IGNORECASE is used
    re_load = re.compile(load_expr, re.IGNORECASE)
    loads = re_load.finditer(filestring)
    # rob.getBehaviour("Base Container").Bases
    cog_dummy = vcMatrix.new()
    inertia_dummy = vcVector.new()
    mass = -1.0
    for load in loads:
      mass = eval(load.group('M'))
      m = vcMatrix.new()
      m.P = vcVector.new(eval(load.group('X')),eval(load.group('Y')),eval(load.group('Z')))
      m.WPR = vcVector.new(eval(load.group('C')),eval(load.group('B')),eval(load.group('A')))
      cog_dummy = m
      inertia_dummy = vcVector.new(eval(load.group('JX')),eval(load.group('JY')),eval(load.group('JZ')))
      print("Imported LOAD_DATA[%s] M %f X  %f  Y  %f  Z  %f  A  %f  B  %f  C  %f Jx %f Jy %f Jz %f" % (cur_tool, mass, eval(load.group('X')), eval(load.group('Y')), eval(load.group('Z')), eval(load.group('A')), eval(load.group('B')), eval(load.group('C')), eval(load.group('JX')), eval(load.group('JY')), eval(load.group('JZ'))))
    
    # setting mass, cog and inertia in robot tool
    try:
      tool_props = tool_container_tools[i].Properties
      mass_prop = None
      cog_prop = None
      inertia_prop = None
      for prop in tool_props:
        if prop.Name == "Mass":
          mass_prop = prop
        elif prop.Name == "CenterOfGravity":
          cog_prop = prop
        elif prop.Name == "Inertia":
          inertia_prop = prop
      
      mass_prop.Value = mass * 1000.0 # mass is defined in Gramm in KUKA.Sim
      cog_prop.Value = cog_dummy
      inertia_prop.Value = inertia_dummy
    
    except IndexError:
      pass # if there are no tool frames, just passing over

  print("------------------------------------------------------------------------------------------------------------------------------------------------------")

  '''Import Home-Pos'''
  num = r'\s*-?\d+\.?\d*[eE]?-?\d*'
  # E6AXIS XHOME={A1 85.0419464,A2 -89.231308,A3 122.052299,A4 -0.0293031298,A5 57.2120895,A6 -5.00842285,E1 -0.500732182874,E2 0.0,E3 0.0,E4 0.0,E5 0.0,E6 0.0}
  # DECL GLOBAL E6AXIS XHOME={A1 0.0,A2 -90.0,A3 90.0,A4 0.0,A5 0.0,A6 0.0,E1 0.0,E2 0.0,E3 0.0,E4 0.0,E5 0.0,E6 0.0}
  home = r'[^;][^; ].*E6AXIS XHOME={A1 (?P<A1>'+num+'),A2 (?P<A2>'+num+'),A3 (?P<A3>'+num+'),A4 (?P<A4>'+num+'),A5 (?P<A5>'+num+'),A6 (?P<A6>'+num+'),E1 (?P<E1>'+num+'),E2 (?P<E2>'+num+'),E3 (?P<E3>'+num+'),E4 (?P<E4>'+num+'),E5 (?P<E5>'+num+'),E6 (?P<E6>'+num+').*'
  #.* allows many sings -
  # for unknown reasons the syntax of E6AXIS XHOME1={a1 0.0,a2 -90.0, ...} in the config.dat sometimes changes to minor case
  # to tackle this re.IGNORECASE is used
  re_home = re.compile(home, re.IGNORECASE)
  homes = re_home.finditer(filestring)
  
  # use tee function from itertools to create copy of iterator homes. Necessary because the check if the iterator is empty or not consumes at least on iterator element.
  homes, homes_copy = tee(homes) # after using tee, only the iterators generated by tee should be used, therefore homes gets re-assigned with the tee generated iterator
  
  # check if iterator holds at least one element -> if iterator is empty we will check for the VSS syntax for a HOME position which is missing the 'X' before HOME
  if not any(homes_copy):
    home = r'[^;][^; ].*E6AXIS HOME={A1 (?P<A1>' + num + '),A2 (?P<A2>' + num + '),A3 (?P<A3>' + num + '),A4 (?P<A4>' + num + '),A5 (?P<A5>' + num + '),A6 (?P<A6>' + num + '),E1 (?P<E1>' + num + '),E2 (?P<E2>'+num+'),E3 (?P<E3>'+num+'),E4 (?P<E4>'+num+'),E5 (?P<E5>'+num+'),E6 (?P<E6>'+num+').*'
    re_home = re.compile(home, re.IGNORECASE)
    homes = re_home.finditer(filestring)
  
  
  for ho in homes:
    #How to get the Child Component
    robotcontroller = rob.findBehavioursByType ( VC_ROBOTCONTROLLER )[0]
    nof_joints = len(robotcontroller.Joints)
    init_values = [0]*12
    for i in range (nof_joints):
      if i < 6:
        joint_name = 'A' + str(i+1)
      else:
        joint_name = 'E' + str(i-5) 
      #print (joint_name)
      robotcontroller.Joints[i].InitialValue = float(ho.group(joint_name))
      init_values[i] = robotcontroller.Joints[i].InitialValue
    
    print("Updated the initial position of robot %s successfully to values of XHOME from config.dat:" % rob.Name)
    print("A1 %f   A2 %f    A3 %f   A4 %f    A5 %f   A6 %f    E1 %f    E2 %f    E3 %f    E4 %f    E5 %f    E6 %f" % ( init_values[0], init_values[1], init_values[2], init_values[3], init_values[4], init_values[5], init_values[6], init_values[7], init_values[8], init_values[9], init_values[10], init_values[11]))
    print('Only the first HOME position will be imported and saved as the initial position of the robot kinematic. Direct import of all HOME positions is not possible!')
  print("------------------------------------------------------------------------------------------------------------------------------------------------------")
  
  rob.update()
  rob.rebuild()
  getApplication().render()
  print('Import Tool/Base finished')

def open_help(para):
  """
  Opens a help file for the add on.
  """
  file_path = ds_helpers.get_script_path(__file__)
  print('Open help file from ' + file_path)
  os.startfile(file_path + '\\Help_File_ImportToolsBases.pdf')

cmd.addState( first_state )
